
#include <stdint.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <stdio.h>


#define DIVISOR   30    // Sampled at 30 MHz / line rate of 1 MHz

static int debug = 0;

static const char _ebcdic_table[256] = {
	'.', '.', '.', '.', '.', '.', '.', '.', '.', '.', '.', '.', '.', '.', '.', '.',
	'.', '.', '.', '.', '.', '.', '.', '.', '.', '.', '.', '.', '.', '.', '.', '.',
	'.', '.', '.', '.', '.', '.', '.', '.', '.', '.', '.', '.', '.', '.', '.', '.',
	'.', '.', '.', '.', '.', '.', '.', '.', '.', '.', '.', '.', '.', '.', '.', '.',
	' ', '.', '.', '.', '.', '.', '.', '.', '.', '.', '.', '.', '<', '(', '+', '|',
	'&', '.', '.', '.', '.', '.', '.', '.', '.', '.', '!', '$', '*', ')', ';', '.',
	'-', '/', '.', '.', '.', '.', '.', '.', '.', '.', '|', ',', '%', '_', '>', '?',
	'.', '.', '.', '.', '.', '.', '.', '.', '.', '`', ':', '#', '@', '\'', '=', '\"',
	'.', 'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', '.', '.', '.', '.', '.', '.',
	'.', 'j', 'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', '.', '.', '.', '.', '.', '.',
	'.', '~', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z', '.', '.', '.', '.', '.', '.',
	'^', '.', '.', '.', '.', '.', '.', '.', '.', '.', '[', ']', '.', '.', '.', '.',
	'{', 'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', '.', '.', '.', '.', '.', '.',
	'}', 'J', 'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', '.', '.', '.', '.', '.', '.',
	'\\', '.', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z', '.', '.', '.', '.', '.', '.',
	'0', '1', '2', '3', '4', '5', '6', '7', '8', '9', '.', '.', '.', '.', '.', '.'
};




static void print_word (uint32_t data)
{
	int i;

	for (i = 0; i < 32; i++)
	{
		if (data & 0x80000000UL)
			printf ("1");
		else
			printf ("0");

		if ((i & 1) == 1)
			printf (" ");

		data = data << 1;
	}

	printf ("\n");

	return;
}



static int decode_frame (uint32_t data, uint16_t *frame_ptr)
{
	uint16_t frame = 0;

	int i;

	for (i = 0; i < 16; i++)
	{
		frame = frame >> 1;

		if (((data & 3) == 3) || ((data & 3) == 0))
			return 0;

		if ((data & 3) == 1)
			frame |= 0x8000UL;

		data  = data  >> 2;
	}

	if (frame_ptr)
		*frame_ptr = frame;

	return 1;
}


typedef struct _st_frame
{
	int start_bit;

	uint8_t byte;
	char pbyte;

	int addr;

	int parity;
	int parity_valid;

	int fill;

	int valid;

} frame_t;

static void validate_frame (uint16_t frame_value, frame_t *frame)
{
	frame->start_bit = (frame_value & 0x8000UL) ? 1 : 0;

	frame->byte = (((frame_value >> 14) & 1) << 0)
	            | (((frame_value >> 13) & 1) << 1)
	            | (((frame_value >> 12) & 1) << 2)
	            | (((frame_value >> 11) & 1) << 3)
	            | (((frame_value >> 10) & 1) << 4)
	            | (((frame_value >>  9) & 1) << 5)
	            | (((frame_value >>  8) & 1) << 6)
	            | (((frame_value >>  7) & 1) << 7);

	frame->pbyte = _ebcdic_table[frame->byte];

	frame->addr = (((frame_value >> 6) & 1) << 0)
	            | (((frame_value >> 5) & 1) << 1)
	            | (((frame_value >> 4) & 1) << 2);

	int parity = 0;

	if ((frame_value >> 14) & 1)
		parity++;

	if ((frame_value >> 13) & 1)
		parity++;

	if ((frame_value >> 12) & 1)
		parity++;

	if ((frame_value >> 11) & 1)
		parity++;

	if ((frame_value >> 10) & 1)
		parity++;

	if ((frame_value >>  9) & 1)
		parity++;

	if ((frame_value >>  8) & 1)
		parity++;

	if ((frame_value >>  7) & 1)
		parity++;

	if ((frame_value >>  6) & 1)
		parity++;

	if ((frame_value >>  5) & 1)
		parity++;

	if ((frame_value >>  4) & 1)
		parity++;

	parity = (parity & 1) ? 0 : 1;

	frame->parity = (frame_value >> 3) & 1;
	frame->parity_valid = (parity ^ frame->parity) ? 0 : 1;

	frame->fill = frame_value & 7;

	frame->valid = 0;

	do
	{
		if (!frame->start_bit)
			break;

		if (!frame->parity_valid)
			break;

		if (frame->fill)
			break;

		frame->valid = 1;

	} while (0);

	return;
}


static uint64_t payload_timecode = 0;
static uint64_t prev_timecode = 0;
static frame_t payload[256];
static int payload_pending = -1;

static int frame_len = 0;         // Current word count in frame


static void _eject_payload (void)
{
	if (payload_timecode == 0)
		return;

	// Frame validation

	int i, bad_frame = 0;

	for (i = 0; i < frame_len; i++)
	{
		if ((!payload[i].start_bit) ||
		    (!payload[i].parity_valid) ||
		    ( payload[i].fill))
		{
			bad_frame = 1;

			break;
		}
	}

	float delta_ms = (payload_timecode - prev_timecode) / 30000.0;

	printf ("%sFRAME +%10.2f ms / %3d byte(s) ",
	        bad_frame ? "BAD " : "    ", delta_ms, frame_len);

	int bt;

	frame_t *frame = payload;

	char badtext[13] = "[";

	for (i = 0; i < frame_len; i++, frame++)
	{
		if (i == 0)
		{
			printf ("[addr=%d] :", frame->addr);
		}

		bt = 1;

		if (!frame->start_bit)
			badtext[bt++] = 'S';

		if (!frame->parity_valid)
			badtext[bt++] = 'P';

		if (frame->fill)
			badtext[bt++] = 'F';

		if (bt == 1)
		{			
			printf (" %02x", frame->byte);
		}
		else
		{
			badtext[bt++] = ']';
			badtext[bt++] = '\0';

			printf (" %02x%s", frame->byte, badtext);

		}
	}


	int response = (payload_pending >= 0) && (delta_ms < 0.07);
	int dump_text = -1;

	if (frame_len)
	{
		printf (" / ");

		if (response)
		{
			switch (payload_pending)
			{
			case 0x10: // Poll
				printf ("Poll RESPONSE :%s%s%s%s%s",
				         (payload[0].byte & 0x80) ? " Busy" : " Not-Busy",
				         (payload[0].byte & 0x40) ? " Parity-Error" : "",
				         (payload[0].byte & 0x20) ? " Update" : "",
				        ((payload[0].byte & 0x0e) == 0x02) ? " Error" :
				        ((payload[0].byte & 0x0e) == 0x04) ? " Invalid-Activate" :
				        ((payload[0].byte & 0x0e) == 0x06) ? " Unknown3" :
				        ((payload[0].byte & 0x0e) == 0x08) ? " Invalid-ID" : 
				        ((payload[0].byte & 0x0e) == 0x0a) ? " Overrun" :
				        ((payload[0].byte & 0x0e) == 0x0c) ? " Invalid-Reg" :
				        ((payload[0].byte & 0x0e) == 0x0e) ? " Power-On" : "",
					    (payload[0].byte & 0x01) ? " Set1" : " Set0");

				for (i = 1; i < frame_len; i++)
				{
					if (payload[i].byte)
						printf (" KeyPress-'%02x'", payload[i].byte);
				}

				break;


			case 0x00: // Read Activate
				printf ("READ RESPONSE");
				dump_text = 0;
				break;


			default:
				printf ("RESPONSE");
				break;
			}

			payload_pending = -1;
		}
		else
		{
			int cmd = payload[0].byte & 0x1f;
			int mod = payload[0].byte >> 5;

			payload_pending = -1;

			switch (cmd)
			{
			case 0x12: // Clear
				printf ("CMD Clear (%d)", mod);
				break;

			case 0x02: // End of Queue
				printf ("CMD End of Queue (%d)", mod);
				break;

			case 0x03: // Insert Character
				printf ("CMD Insert Character (%d)", mod);
				break;

			case 0x15: // Load Address Counter
				printf ("CMD Load Address Counter (%d)", mod);

				if (frame_len >= 3)
				{
					int offset = (payload[1].byte << 8) | payload[2].byte;

					int row = (offset / 80) + 1;
					int col = (offset % 80) + 1;

					printf (" / ROW=%d COL=%d", row, col);
				}

				break;

			case 0x17: // Load Cursor Register
				printf ("CMD Load Cursor Register (%d)", mod);

				if (frame_len >= 3)
				{
					int offset = (payload[1].byte << 8) | payload[2].byte;

					int row = (offset / 80) + 1;
					int col = (offset % 80) + 1;

					printf (" / ROW=%d COL=%d", row, col);
				}

				break;

			case 0x07: // Load Reference Counter
				printf ("CMD Load Reference Counter (%d)", mod);

				if (frame_len >= 3)
				{
					int offset = (payload[1].byte << 8) | payload[2].byte;

					int row = (offset / 80) + 1;
					int col = (offset % 80) + 1;

					printf (" / ROW=%d COL=%d", row, col);
				}

				break;

			case 0x06: // Move Data
				printf ("CMD Move Data (%d)", mod);
				break;

			case 0x10: // Poll
				printf ("CMD Poll (%d)", mod);
				payload_pending = cmd;
				break;

			case 0x00: // Read Activate
				printf ("CMD Read Activate (%d)", mod);
				payload_pending = cmd;
				break;

			case 0x01: // Write Activate
				printf ("CMD Write Activate (%d)", mod);
				dump_text = 1;
				break;

			case 0x1a: // Read Between Limits
				printf ("CMD Read Between Limits (%d)", mod);
				payload_pending = cmd;
				break;

			case 0x08: // Read Data
				printf ("CMD Read Data (%d)", mod);
				payload_pending = cmd;
				break;

			case 0x1c: // Read Registers
				printf ("CMD Read Registers (%d)", mod);
				payload_pending = cmd;
				break;

			case 0x0a: // Read to End of Line
				printf ("CMD Read to End of Line (%d)", mod);
				payload_pending = cmd;
				break;
#if 0
			case 0x02: // Reset
				printf ("CMD Reset (%d)", mod);
				break;
#endif
			case 0x04: // Search Next Null **
				printf ("CMD Search Next Null (%d)", mod);
				payload_pending = cmd;
				break;

			case 0x13: // Set Mode
				printf ("CMD Set Mode (%d)", mod);
				break;

			case 0x05: // Write Control Data
				printf ("CMD Write Control Data (%d)", mod);
				break;

			case 0x11: // Write Data and Load Cursor
				printf ("CMD Write Data and Load Cursor (%d)", mod);
				break;

			case 0x1d: // Write Immediate Data
				printf ("CMD Write Immediate Data (%d)", mod);
				dump_text = 1;
				break;

			default:
				printf ("Unknown CMD %x (%d)", cmd, mod);
				break;
			}
		}
	}

	if ((dump_text >= 0) && (frame_len >= dump_text))
	{
		frame_t *frame = payload;

		printf (" \"");

		for (i = dump_text; i < frame_len; i++, frame++)
			printf ("%c", frame->pbyte);

		printf ("\"");
	}

	printf ("\n");


	prev_timecode = payload_timecode;

	return;
}







static uint64_t timecode = 0;     // Total bytes (/DIVISOR for micro-sec)
static int zero_count = 0;        // Return to idle bit count
static int line_search = 1;       // Looking for start bit if true
static int tick_count;            // Clock ticks in current half-bit
static uint32_t line_word;        // half-bit shift register
static int ignore_bits = 0;       // Grace perious after frames (unused)


static void _proc_bit (uint8_t data)
{
	timecode++;


	int ignore = ignore_bits;

	if (ignore_bits)
		ignore_bits--;


	// Global line idle check for return to search

	if (data)
		zero_count = 0;
	else
		zero_count++;

	if (zero_count > (6 * DIVISOR))
	{
		// Return to searching for start bit

		if (frame_len)
		{
			_eject_payload ();

			frame_len = 0;
		}

		if (!line_search && debug)
		{
			printf ("RETURN TO IDLE : ");

			print_word (line_word);

			printf ("\n");
		}

		line_search = 1;
		zero_count  = 0;

		return;
	}

	if (ignore)
		return;


	if (line_search) // looking for start
	{
		if (data)
		{
			line_search = 0;
			tick_count  = 1;
			line_word   = 0;

			if (debug)
				printf ("START BIT\n");
		}

		return;
	}


	// Process shift register values

	tick_count++;

	if (tick_count == (DIVISOR / 4))
	{
		line_word = (line_word << 1) | (data ? 1 : 0);

		return;
	}

	if (tick_count == (DIVISOR / 2))
	{
		tick_count = 0;

		/*
		 * Continually compare current shift register
		 * value for a 8-bit framing start pattern in the
		 * lower half in the form of:
		 *
		 *    01 01 01 01 01 000 111 
		 *
		 * This will signal a reset of the frame's word
		 * count.  Otherwise this is a word in the frame
		 * and we should wait until the start bit moves
		 * into the most significant half-bit positions.
		 */

		if ((line_word & 0x0000ffffUL) == 0x5547)
		{
			if (frame_len)
			{
				_eject_payload ();

				frame_len = 0;
			}

			if (debug)
				printf ("FRAME SYNC\n");

			line_search = 1;
//			ignore_bits = DIVISOR / 4;

			payload_timecode = timecode - (8 * DIVISOR);
		}
		else if (line_word & 0x40000000UL) // MS half-bit = 1
		{
			line_search = 1;
//			ignore_bits = DIVISOR / 4;

			uint16_t frame_value = 0;

			if (!decode_frame (line_word, &frame_value))
			{
				if (line_word && debug)
				{
					printf ("BAD DECODE ");

					print_word (line_word);
				}

				return;
			}

			frame_t frame;

			validate_frame (frame_value, &frame);

			memcpy (&payload[frame_len], &frame, sizeof (frame_t));
			frame_len++;

			if (debug)
			{
				printf ("START %d / BYTE %02x('%c') / ADDR %d / PARITY %s / FILL %x [%04x] - ",
					   frame.start_bit,
				        frame.byte, frame.pbyte,
				        frame.addr,
					   frame.parity_valid ? "GOOD" : "BAD",
				        frame.fill, frame_value);

				print_word (line_word);
			}

			if (frame.valid && (frame.addr == 7))
			{
				_eject_payload ();

				frame_len = 0;
			}
		}
	}

	return;
}




#include <zlib.h>


int main (int argc, char *argv[])
{
//	gzFile file = gzopen ("info.comp.gz", "r");
	gzFile file = gzopen ("3179.comp.gz", "r");

	uint8_t buffer[512];

	while (1)
	{
		int len = gzread (file, buffer, sizeof (buffer));

		if (len <= 0)
			break;

		int i;

		for (i = 0; i < len; i++)
		{
			_proc_bit (((buffer[i] >> 0) & 1) ? 1 : 0);
			_proc_bit (((buffer[i] >> 1) & 1) ? 1 : 0);
			_proc_bit (((buffer[i] >> 2) & 1) ? 1 : 0);
			_proc_bit (((buffer[i] >> 3) & 1) ? 1 : 0);
			_proc_bit (((buffer[i] >> 4) & 1) ? 1 : 0);
			_proc_bit (((buffer[i] >> 5) & 1) ? 1 : 0);
			_proc_bit (((buffer[i] >> 6) & 1) ? 1 : 0);
			_proc_bit (((buffer[i] >> 7) & 1) ? 1 : 0);
		}
	}

	gzclose (file);

	return 0;
}
